'''
MIT License

Copyright (c) 2021 Qualichat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import datetime
import os
import random
import re
from typing import List, Dict

import emojis

from .abc import Message as BaseMessage
from .enums import _get_period, _get_sub_period


time_format = r'%d/%m/%y %H:%M:%S'

def _parse_time(string: str) -> datetime.datetime:
    return datetime.datetime.strptime(string, time_format)


_path = os.path.dirname(__file__)
_books_path = os.path.join(_path, 'books.txt')

with open(_books_path) as f:
    __books__ = f.read().split('\n')


def _get_random_name() -> str:
    name = random.choice(__books__)
    # Remove the book from the list so there is no risk that two 
    # actors have the same display name.
    __books__.remove(name)
    return name.strip()


URL_REGEX = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*(),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
EMAIL_REGEX = re.compile(r'\b[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,}\b', re.I)
NUMBERS_REGEX = re.compile(r'\d+')
LAUGHS_REGEX = re.compile(r'\s((?:he|ha|hi|hu){2,}|(?:hh){1,}|(?:ja|je|ka){2,}|(?:kk|rs){1,})', re.I)
EXCLAMATION_MARK_REGEX = re.compile(r'!+')
QUESTION_MARK_REGEX = re.compile(r'\?+')


class Actor:
    '''Represents an actor in the chat.

    Attributes
    -----------
    display_name: :class:`str`
        A representative name for this actor, this name is 
        not necessarily the user's real name.
    messages: List[:class:`Message`]
        A list containing all the messages that this user sent 
        in the chat.
    '''

    __slots__ = ('__contact_name__', 'messages', 'display_name')

    def __init__(self, contact_name: str):
        self.__contact_name__ = contact_name
        self.messages = []
        self.display_name = _get_random_name()

    def __repr__(self):
        return f'<Actor display_name={self.display_name!r} messages={len(self.messages)}>'


class Message(BaseMessage):
    '''Represents a message sent in the chat.'''

    __slots__ = ('actor', 'content', 'created_at', 'period', 'sub_period')

    def __init__(self, *, actor: Actor, content: str, created_at: str):
        self.actor = actor
        self.content = content
        self.created_at = _parse_time(created_at)
        self.period = _get_period(self.created_at)
        self.sub_period = _get_sub_period(self.created_at)

    def __repr__(self):
        return '<Message actor={0.actor} created_at={0.created_at}>'.format(self)

    @property
    def emojis(self) -> List[str]:
        return list(emojis.iter(self.content))

    @property
    def urls(self) -> List[str]:
        return URL_REGEX.findall(self.content)

    @property
    def emails(self) -> List[str]:
        return EMAIL_REGEX.findall(self.content)

    @property
    def numbers(self) -> List[str]:
        content = self.content

        for url in self.urls:
            content = content.replace(url, '')

        return NUMBERS_REGEX.findall(content)

    @property
    def laughs(self) -> List[str]:
        content = self.content

        for url in self.urls:
            content = content.replace(url, '')

        return LAUGHS_REGEX.findall(content)

    @property
    def marks(self) -> Dict[str, List[str]]:
        marks = {
            'question_marks': self.question_marks,
            'exclamation_mark': self.exclamation_marks
        }

        return marks

    @property
    def exclamation_marks(self) -> List[str]:
        content = self.content

        for url in self.urls:
            content = content.replace(url, '')

        return EXCLAMATION_MARK_REGEX.findall(content)

    @property
    def question_marks(self) -> List[str]:
        content = self.content

        for url in self.urls:
            content = content.replace(url, '')

        return QUESTION_MARK_REGEX.findall(content)

    @property
    def liquid(self) -> str:
        content = self.content

        for url in self.urls:
            content = content.replace(url, '')

        for email in self.emails:
            content = content.replace(email, '')

        for emoji in self.emojis:
            content = content.replace(emoji, '')

        return content

    @property
    def pure_text(self) -> str:
        content = self.liquid

        for laugh in self.laughs:
            content = content.replace(laugh, '')

        for exclamation_mark in self.exclamation_marks:
            content = content.replace(exclamation_mark, '')

        for question_mark in self.question_marks:
            content = content.replace(question_mark, '')

        for number in self.numbers:
            content = content.replace(number, '')

        return content


class SystemMessage(BaseMessage):
    '''Represents a system message sent in the chat.
    All of these messages were sent by the WhatsApp app automatically, 
    that means that no actor sent them.
    '''

    __slots__ = ('content', 'created_at', 'period', 'sub_period')
    
    def __init__(self, *, content: str, created_at: str):
        self.content = content
        self.created_at = _parse_time(created_at)
        self.period = _get_period(self.created_at)
        self.sub_period = _get_sub_period(self.created_at)

    def __repr__(self):
        return '<SystemMessage created_at={0.created_at}>'.format(self)
